/*
 * Author: Kilian
 * Date: 2025-04-20
 * Description: This file contains an implementation of the ISO-TP protocol
 *              with both transmission and reception capabilities.
 *              
 * References:
 * https://munich.dissec.to/kb/chapters/isotp/isotp.html#protocol-basics-iso-15765-2
 * https://www.kernel.org/doc/html/next//networking/iso15765-2.html
 * https://en.wikipedia.org/wiki/ISO_15765-2
 * https://ramn.readthedocs.io/en/latest/userguide/isotp_tutorial.html
 */
#include "isotp.h"
#include "isotp_user_defs.h"

#include <stdint.h>
#include <stddef.h>
#include <string.h>

#define MAX_CAN_DATA_SIZE 8
#define MAX_ISOTP_DATA_SIZE_MESSAGE 4095
#define MAX_ISOTP_DATA_SINGLE 7
#define MAX_ISOTP_DATA_FIRST 6
#define MAX_ISOTP_DATA_CONSECUTIVE 7

// check endianness
static_assert(__BYTE_ORDER__ == __ORDER_LITTLE_ENDIAN__, "Endianness is not little endian, this code is not portable!");

typedef union [[gnu::packed]] {
  struct {
    uint8_t header;
    uint8_t data[MAX_ISOTP_DATA_SINGLE];
  } single;
  struct {
    uint8_t header_1;
    uint8_t header_2;
    uint8_t data[MAX_ISOTP_DATA_FIRST];
  } first;
  struct {
    uint8_t header;
    uint8_t data[MAX_ISOTP_DATA_CONSECUTIVE];
  } consecutive;
} isotp_message;

enum {
    ISOTP_SINGLE_FRAME = 0 << 4,
    ISOTP_FIRST_FRAME = 1 << 4,
    ISOTP_CONSECUTIVE_FRAME = 2 << 4,
    ISOTP_FLOW_CONTROL = 3 << 4
};

struct isotp_message_state {
#if ISOTP_RX
    uint8_t * data;
#else
    const uint8_t * data;
#endif
    uint32_t timestamp;
    uint16_t remaining;
#if ISOTP_RX
    uint16_t data_len;
#endif
    isotp_conn_id_t conn_id;  // Store connection ID instead of searching
    enum : uint8_t {
        ISOTP_NEW_MESSAGE = 0,
        ISOTP_WAITING_FLOW_CONTROL = 1,
        ISOTP_READY = 2,
        ISOTP_DONE = 3,
        ISOTP_FAILED = 4
    } state;
    struct {
        enum : uint8_t {
            ISOTP_CONTINUE = 0,
            ISOTP_WAIT = 1,
            ISOTP_ABORT = 2
        } state;
        uint8_t block_size;
        uint8_t st_min; // in ms, we round up to 1 ms if sub 1 ms
        uint8_t st_min_counter;
    } flow_control;
    uint8_t seq_n;
};

static struct isotp_message_state tx_queue[MAX_ISOTP_MESSAGES_IN_FLIGHT] = {[0 ... MAX_ISOTP_MESSAGES_IN_FLIGHT - 1] = {.state=ISOTP_DONE }};

#if ISOTP_RX
static struct isotp_message_state rx_queue[MAX_ISOTP_MESSAGES_IN_FLIGHT] = {[0 ... MAX_ISOTP_MESSAGES_IN_FLIGHT - 1] = {.state=ISOTP_DONE }};
#endif

#define INVALID_CAN_ID 0xFFFF
#define INVALID_CONN_ID MAX_ISOTP_CONNECTIONS

static struct {
    uint16_t them;
    uint16_t us;
    isotp_flags_t flags;
} connections[MAX_ISOTP_CONNECTIONS] = {
    [0 ... MAX_ISOTP_CONNECTIONS - 1] = {.them=INVALID_CAN_ID, .us=INVALID_CAN_ID, .flags=0}
};

static uint32_t last_called;

isotp_status_t isotp_init() {
    // platform-specific initialization
    last_called = isotp_get_time() - 100;
    return ISOTP_OK;
}

#if ISOTP_RX
#define DATA_TYPE uint8_t *
#else
#define DATA_TYPE const uint8_t *
#endif

/**
 * Helper function to check if a connection has a specific flag set
 */
static inline bool check_flag(isotp_conn_id_t conn_id, isotp_flags_t flags) {
    return (conn_id < MAX_ISOTP_CONNECTIONS &&
            (connections[conn_id].flags & flags) == flags);
}

// Internal helper structure to return both status and index
typedef struct {
    isotp_status_t status;
    size_t index;  // Index of available slot if status is ISOTP_OK
} isotp_validation_result_t;

// Internal helper function to validate message parameters and find available slot
ACCESS(none, 2, 3)
static isotp_validation_result_t isotp_validate_message(isotp_conn_id_t conn_id, const uint8_t *data, size_t datalen) {
    isotp_validation_result_t result = {.status = ISOTP_ERROR, .index = 0};
    
    if (datalen > MAX_ISOTP_DATA_SIZE_MESSAGE) {
        result.status = ISOTP_DATA_TOO_LONG;
        return result;
    }

    if (data == NULL || datalen == 0) {
        result.status = ISOTP_INVALID_PARAMETER;
        return result;
    }

    // Validate connection ID
    if (conn_id >= MAX_ISOTP_CONNECTIONS || connections[conn_id].them == INVALID_CAN_ID) {
        result.status = ISOTP_CONNECTION_NOT_FOUND;
        return result;
    }

    // Check if message with same connection ID is already in flight
    for (size_t i = 0; i < MAX_ISOTP_MESSAGES_IN_FLIGHT; i++) {
        if (!(tx_queue[i].state == ISOTP_DONE || tx_queue[i].state == ISOTP_FAILED) && tx_queue[i].conn_id == conn_id) {
            result.status = ISOTP_MESSAGE_ALREADY_IN_FLIGHT;
            return result;
        }
    }

    // Find available slot
    for (size_t i = 0; i < MAX_ISOTP_MESSAGES_IN_FLIGHT; i++) {
        if (tx_queue[i].state == ISOTP_DONE || tx_queue[i].state == ISOTP_FAILED) {
            result.status = ISOTP_OK;
            result.index = i;
            return result;
        }
    }

    result.status = ISOTP_TOO_MANY_MESSAGES;
    return result;
}

isotp_status_t isotp_add_message(isotp_conn_id_t conn_id, const uint8_t * data, size_t datalen) {
    const isotp_validation_result_t result = isotp_validate_message(conn_id, data, datalen);
    
    if (result.status == ISOTP_OK) {
        // Add message to the queue at the found index
        tx_queue[result.index] = (struct isotp_message_state) {
            .data = (DATA_TYPE) data, 
            .conn_id = conn_id,
            .state = ISOTP_NEW_MESSAGE, 
            .seq_n = 1,
            .remaining = datalen
        };
    }
    
    return result.status;
}

isotp_status_t isotp_try_add_message(isotp_conn_id_t conn_id, const uint8_t * data, size_t datalen) {
    // Just validate without adding to the queue
    const isotp_validation_result_t result = isotp_validate_message(conn_id, data, datalen);
    return result.status;
}

/* Macro to check CAN transmission result and handle errors */
#define TRY_CAN_TRANSMIT(id, data, size) \
    if (!isotp_transmit(id, data, size)) { \
        tx_queue[i].state = ISOTP_FAILED; \
        return ISOTP_CAN_TRANSMIT_ERROR; \
    }

isotp_status_t isotp_update() {
    const uint32_t now = isotp_get_time();
    const uint32_t delta = now - last_called;
    for (size_t i = 0; i < MAX_ISOTP_MESSAGES_IN_FLIGHT; i++) {

        isotp_message msg = {};
        auto conn = connections[tx_queue[i].conn_id];

        switch (tx_queue[i].state) {
            case ISOTP_NEW_MESSAGE:
                if (tx_queue[i].remaining <= MAX_ISOTP_DATA_SINGLE) {
                    msg.single.header = ISOTP_SINGLE_FRAME | (tx_queue[i].remaining & 0x0F);
                    memcpy(msg.single.data, tx_queue[i].data, tx_queue[i].remaining);
                    #if ISOTP_PADDING
                    memset(msg.single.data + tx_queue[i].remaining, ISOTP_PADDING, MAX_ISOTP_DATA_SINGLE - tx_queue[i].remaining);
                    TRY_CAN_TRANSMIT(conn.us, (const uint8_t *)&msg, MAX_CAN_DATA_SIZE);
                    #else
                    TRY_CAN_TRANSMIT(conn.us, (const uint8_t *)&msg, tx_queue[i].remaining + 1);
                    #endif
                    tx_queue[i].remaining = 0;
                    tx_queue[i].state = ISOTP_DONE;
                } else {
                    msg.first.header_1 = ISOTP_FIRST_FRAME | ((tx_queue[i].remaining >> 8) & 0x0F);
                    msg.first.header_2 = tx_queue[i].remaining & 0xFF;
                    memcpy(msg.first.data, tx_queue[i].data, MAX_ISOTP_DATA_FIRST);
                    TRY_CAN_TRANSMIT(conn.us, (const uint8_t *)&msg, MAX_CAN_DATA_SIZE);
                    tx_queue[i].remaining -= MAX_ISOTP_DATA_FIRST;
                    tx_queue[i].data += MAX_ISOTP_DATA_FIRST;
                    
                    // Check if we're in broadcast mode - don't wait for flow control
                    if (check_flag(tx_queue[i].conn_id, ISOTP_FLAGS_BROADCAST)) {
                        // In broadcast mode, assume we can send directly
                        tx_queue[i].flow_control.state = ISOTP_CONTINUE;
                        tx_queue[i].flow_control.block_size = 0; // No block size limit
                        tx_queue[i].flow_control.st_min = 0;     // No separation time
                        tx_queue[i].flow_control.st_min_counter = 0;
                        tx_queue[i].state = ISOTP_READY;
                        goto ISOTP_SEND_AGAIN; // Send all frames immediately
                    } else {
                        tx_queue[i].timestamp = now; // Set timestamp when entering WAITING_FLOW_CONTROL state
                        tx_queue[i].state = ISOTP_WAITING_FLOW_CONTROL;
                    }
                }
                break;
            case ISOTP_READY:
            ISOTP_SEND_AGAIN:
                if (tx_queue[i].flow_control.st_min_counter <= delta) {
                    tx_queue[i].flow_control.st_min_counter = tx_queue[i].flow_control.st_min;
                } else {
                    tx_queue[i].flow_control.st_min_counter -= delta;
                    break;
                }
                if (tx_queue[i].flow_control.block_size == 1) {
                    tx_queue[i].state = ISOTP_WAITING_FLOW_CONTROL;
                }
                if (tx_queue[i].flow_control.block_size != 0) {
                    tx_queue[i].flow_control.block_size--;
                }

                msg.consecutive.header = ISOTP_CONSECUTIVE_FRAME | (tx_queue[i].seq_n & 0x0F);
                size_t bytes_to_send = tx_queue[i].remaining <= MAX_ISOTP_DATA_CONSECUTIVE ? tx_queue[i].remaining : MAX_ISOTP_DATA_CONSECUTIVE;
                memcpy(msg.consecutive.data, tx_queue[i].data, bytes_to_send);
                
                #if ISOTP_PADDING
                if (bytes_to_send < MAX_ISOTP_DATA_CONSECUTIVE) {
                    memset(msg.consecutive.data + bytes_to_send, ISOTP_PADDING, MAX_ISOTP_DATA_CONSECUTIVE - bytes_to_send);
                    TRY_CAN_TRANSMIT(conn.us, (const uint8_t *)&msg, MAX_CAN_DATA_SIZE);
                } else {
                    TRY_CAN_TRANSMIT(conn.us, (const uint8_t *)&msg, MAX_CAN_DATA_SIZE);
                }
                #else
                TRY_CAN_TRANSMIT(conn.us, (const uint8_t *)&msg, bytes_to_send + 1);
                #endif
                
                tx_queue[i].data += bytes_to_send;
                tx_queue[i].remaining -= bytes_to_send;
                tx_queue[i].seq_n++;
                
                if (tx_queue[i].remaining == 0) {
                    tx_queue[i].state = ISOTP_DONE;
                } 
                
                if (tx_queue[i].state == ISOTP_READY) {
                    if (tx_queue[i].flow_control.st_min == 0) { // no delay needed, and we are not waiting for flow control -> send again
                        goto ISOTP_SEND_AGAIN;
                    }
                }
                break;
            case ISOTP_WAITING_FLOW_CONTROL:
                if (now - tx_queue[i].timestamp > ISOTP_FC_WAIT_TIMEOUT) {
                    tx_queue[i].state = ISOTP_FAILED; // Timeout handling
                }
                break;
            case ISOTP_DONE:
            case ISOTP_FAILED:
                break;
        }
    }

    #if ISOTP_RX
    for (size_t i = 0; i < MAX_ISOTP_MESSAGES_IN_FLIGHT; i++) {
        if (rx_queue[i].state == ISOTP_READY) {
            if (now - rx_queue[i].timestamp > ISOTP_RX_TIMEOUT) {
                rx_queue[i].state = ISOTP_FAILED;
            }
        }
    }
    #endif
    last_called = now;
    return ISOTP_OK;
}

int isotp_add_connection(uint16_t them, uint16_t us, isotp_flags_t flags) {
    if (flags > (ISOTP_FLAGS_BROADCAST | ISOTP_FLAGS_LISTEN_ONLY)) {
        return ISOTP_INVALID_FLAGS;
    }
    
    for (size_t i = 0; i < MAX_ISOTP_CONNECTIONS; i++) {
        if (connections[i].them == INVALID_CAN_ID) {
            connections[i].them = them;
            connections[i].us = us;
            connections[i].flags = flags;
            return i;  // Return the connection ID on success
        }
    }
    
    return ISOTP_TOO_MANY_CONNECTIONS;
}

enum FC_status : uint8_t {
    ISOTP_FC_CTS = 0,
    ISOTP_FC_WAIT = 1,
    ISOTP_FC_ABORT = 2
};

// Function prototypes for internal functions
#if ISOTP_RX
static isotp_status_t isotp_send_fc(isotp_conn_id_t conn_id, enum FC_status status, uint8_t block_size, uint8_t st_min);
static isotp_status_t isotp_handle_single_frame(isotp_conn_id_t conn_id, const uint8_t * data, size_t datalen);
static isotp_status_t isotp_handle_first_frame(isotp_conn_id_t conn_id, const uint8_t * data, size_t datalen);
static isotp_status_t isotp_handle_consecutive_frame(isotp_conn_id_t conn_id, const uint8_t * data, size_t datalen);
#endif
static isotp_status_t isotp_handle_flow_control(isotp_conn_id_t conn_id, const uint8_t * data, size_t datalen);

#if ISOTP_RX
static isotp_status_t isotp_send_fc(isotp_conn_id_t conn_id, enum FC_status status, uint8_t block_size, uint8_t st_min) {
    // Check if we're in listen-only mode first
    if (check_flag(conn_id, ISOTP_FLAGS_LISTEN_ONLY)) {
        return ISOTP_OK; // Silently succeed but don't actually send anything
    }

    isotp_message msg = {};
    msg.single.header = ISOTP_FLOW_CONTROL | (status & 0x0F);
    msg.single.data[0] = block_size;
    msg.single.data[1] = st_min;

    #if ISOTP_PADDING
    memset(msg.single.data + 2, ISOTP_PADDING, MAX_ISOTP_DATA_SINGLE - 2);
    if (!isotp_transmit(connections[conn_id].us, (const uint8_t *)&msg, MAX_CAN_DATA_SIZE)) {
        return ISOTP_CAN_TRANSMIT_ERROR;
    }
    #else
    if (!isotp_transmit(connections[conn_id].us, (const uint8_t *)&msg, 3)) {
        return ISOTP_CAN_TRANSMIT_ERROR;
    }
    #endif
   
    return ISOTP_OK;
}

static isotp_status_t isotp_handle_first_frame(isotp_conn_id_t conn_id, const uint8_t * data, size_t datalen) {
    if (datalen < 3) {
        return ISOTP_INVALID_FRAME;
    }

    const uint16_t length = ((data[0] & 0x0F) << 8) | data[1];
    if (length > MAX_ISOTP_DATA_SIZE_MESSAGE) {
        return ISOTP_INVALID_FRAME;
    }

    for (size_t i = 0; i < MAX_ISOTP_MESSAGES_IN_FLIGHT; i++) {
        if (rx_queue[i].state == ISOTP_DONE || rx_queue[i].state == ISOTP_FAILED) {
            uint8_t* rx_buffer = isotp_get_rx_buffer(length);
            if (rx_buffer == NULL) {
                isotp_send_fc(conn_id, ISOTP_FC_ABORT, 0, 0); // Abort
                return ISOTP_BUFFER_CREATION_ERROR;
            }

            rx_queue[i] = (struct isotp_message_state) {
                .data = rx_buffer, 
                .conn_id = conn_id,  // Store connection ID
                .remaining = length, 
                .data_len = length, 
                .state = ISOTP_NEW_MESSAGE,
                .seq_n = 1 // Initialize sequence number
            };
            memcpy(rx_queue[i].data, data + 2, MAX_ISOTP_DATA_FIRST);
            rx_queue[i].remaining -= MAX_ISOTP_DATA_FIRST;
            rx_queue[i].data += MAX_ISOTP_DATA_FIRST;
            rx_queue[i].timestamp = isotp_get_time();
            rx_queue[i].state = ISOTP_READY;
            
            isotp_send_fc(conn_id, ISOTP_FC_CTS, 0, ISOTP_RX_MIN_ST); // Send flow control frame
            return ISOTP_OK;
        }
    }
    
    isotp_send_fc(conn_id, ISOTP_FC_ABORT, 0, 0); // Abort
    return ISOTP_TOO_MANY_MESSAGES;
}

static isotp_status_t isotp_handle_consecutive_frame(isotp_conn_id_t conn_id, const uint8_t * data, size_t datalen) {
    if (datalen < 2) {
        return ISOTP_INVALID_FRAME;
    }

    const uint8_t seq_n = data[0] & 0x0F;

    for (size_t i = 0; i < MAX_ISOTP_MESSAGES_IN_FLIGHT; i++) {
        if (rx_queue[i].state == ISOTP_DONE || rx_queue[i].state == ISOTP_FAILED || rx_queue[i].conn_id != conn_id) {
            continue;
        }

        if (rx_queue[i].seq_n != seq_n) {
            isotp_send_fc(conn_id, ISOTP_FC_ABORT, 0, 0); // Abort
            rx_queue[i].state = ISOTP_FAILED;
            isotp_free_rx_buffer(rx_queue[i].data - rx_queue[i].data_len + rx_queue[i].remaining); // Free the buffer
            return ISOTP_INVALID_SEQUENCE; // Sequence number mismatch
        }

        rx_queue[i].seq_n++;
        rx_queue[i].seq_n &= 0x0F; // Wrap around sequence number
        
        const size_t bytes_to_copy = rx_queue[i].remaining <= MAX_ISOTP_DATA_CONSECUTIVE ? rx_queue[i].remaining : MAX_ISOTP_DATA_CONSECUTIVE;
        memcpy(rx_queue[i].data, data + 1, bytes_to_copy);
        rx_queue[i].data += bytes_to_copy;
        rx_queue[i].remaining -= bytes_to_copy;
        rx_queue[i].timestamp = isotp_get_time(); // Update timestamp for timeout handling

        if (rx_queue[i].remaining == 0) {
            rx_queue[i].state = ISOTP_DONE;
            isotp_on_message_received(connections[conn_id].them, rx_queue[i].data - rx_queue[i].data_len, rx_queue[i].data_len);
            isotp_free_rx_buffer(rx_queue[i].data - rx_queue[i].data_len); // Free the buffer after message is received
        }

        return ISOTP_OK;
    }

    isotp_send_fc(conn_id, ISOTP_FC_ABORT, 0, 0); // Abort
    return ISOTP_INVALID_MESSAGE; // No matching message found
}

static isotp_status_t isotp_handle_single_frame(isotp_conn_id_t conn_id, const uint8_t * data, size_t datalen) {
    if (datalen < 2) {
        return ISOTP_INVALID_FRAME;
    }

    const uint8_t length = data[0] & 0x0F;
    if (length > MAX_ISOTP_DATA_SINGLE) {
        return ISOTP_INVALID_FRAME;
    }

    if (datalen < (unsigned) (length + 1)) {
        return ISOTP_INVALID_FRAME;
    }

    isotp_on_message_received(connections[conn_id].them, data + 1, length); // Send the correct length of data
    return ISOTP_OK;
}
#endif // ISOTP_RX

static isotp_status_t isotp_handle_flow_control(isotp_conn_id_t conn_id, const uint8_t * data, size_t datalen) {
    if (datalen < 3) {
        return ISOTP_INVALID_FRAME;
    }

    for (size_t i = 0; i < MAX_ISOTP_MESSAGES_IN_FLIGHT; i++) {
        if (tx_queue[i].state == ISOTP_DONE || tx_queue[i].state == ISOTP_FAILED || tx_queue[i].conn_id != conn_id) {
            continue;
        }

        const uint8_t status = data[0] & 0x0F;
        const uint8_t block_size = data[1];
        const uint8_t st_min = data[2];

        switch (status) {
            case ISOTP_FC_CTS: // Continue
                tx_queue[i].flow_control.state = ISOTP_CONTINUE;
                break;
            case ISOTP_FC_WAIT: // Wait
                tx_queue[i].flow_control.state = ISOTP_WAIT;
                tx_queue[i].state = ISOTP_WAITING_FLOW_CONTROL;
                tx_queue[i].timestamp = isotp_get_time(); // Update timestamp for timeout handling
                return ISOTP_OK;
            case ISOTP_FC_ABORT: // Abort
                tx_queue[i].flow_control.state = ISOTP_ABORT;
                tx_queue[i].state = ISOTP_FAILED;
                return ISOTP_OK;
            default:
                return ISOTP_INVALID_FRAME;
        }

        tx_queue[i].flow_control.block_size = block_size;
        tx_queue[i].flow_control.st_min = st_min > 0x7F ? 1 : st_min; // round up to 1 ms
        tx_queue[i].flow_control.st_min_counter = st_min > 0x7F ? 1 : st_min;
        tx_queue[i].state = ISOTP_READY;

        return ISOTP_OK;
    }
    return ISOTP_INVALID_PARAMETER;
}

isotp_status_t isotp_handle_incoming(uint16_t id, const uint8_t * data, size_t datalen) {
    if (data == NULL) {
        return ISOTP_INVALID_PARAMETER;
    }

    // Find the matching connection ID
    isotp_conn_id_t conn_id = INVALID_CONN_ID;
    for (size_t i = 0; i < MAX_ISOTP_CONNECTIONS; i++) {
        if (connections[i].them == id) {
            conn_id = i;
            break;
        }
    }

    if (conn_id == INVALID_CONN_ID) {
        return ISOTP_CONNECTION_NOT_FOUND;
    }

    const uint8_t type = data[0] & 0xF0;

    switch (type) {
#if ISOTP_RX
        case ISOTP_SINGLE_FRAME:
            return isotp_handle_single_frame(conn_id, data, datalen);
        case ISOTP_FIRST_FRAME:
            return isotp_handle_first_frame(conn_id, data, datalen);
        case ISOTP_CONSECUTIVE_FRAME:
            return isotp_handle_consecutive_frame(conn_id, data, datalen);
#endif
        case ISOTP_FLOW_CONTROL:
            return isotp_handle_flow_control(conn_id, data, datalen);
        default:
            return ISOTP_INVALID_FRAME;
    }
}

