/*
 * swo_log_backend.c
 *
 * Created on: 21.04.2025
 * Author: Kilian
 *
 * SWO backend implementation for the logging system
 */

#include "swo_log_backend.h"
#include "log.h"
#include "swo_util.h" // Added include for the new utility functions

/* Global variables */
#if USE_CHANNEL_MASK_VARIABLE
volatile uint32_t MASK_VARIABLE = 0b11111;  // LOG_NOISY off by default
#endif

static bool swo_backend_registered = false;

/* SWO backend functions */
static void swo_backend_init(void);
static bool swo_backend_is_enabled(log_level_t level);
static void swo_backend_write(const log_message_t* message);
static void swo_backend_flush(void);
static void swo_backend_clear(void);

/* SWO backend definition */
static const log_backend_t swo_backend = {
  .name = "SWO",
  .init = swo_backend_init,
  .is_enabled = swo_backend_is_enabled,
  .write = swo_backend_write,
  .flush = swo_backend_flush,
  .clear = swo_backend_clear
};

/* Return the SWO backend definition */
const log_backend_t* swo_log_get_backend(void) {
    return &swo_backend;
}

/* Check if the SWO backend is registered */
bool swo_log_is_backend_registered(void) {
    return swo_backend_registered;
}

/* Initialize the SWO backend and register it */
void swo_log_backend_init(void) {
    /* Register the backend */
    swo_backend_registered = log_register_backend(&swo_backend);
}

/* SWO backend initialization */
static void swo_backend_init(void) {
    /* SWO is initialized by the debugger, nothing to do here */
}

/* Check if the SWO backend is enabled for the given log level */
static bool swo_backend_is_enabled(log_level_t level) {
    unsigned int channel_to_check;
#if USE_MULTIPLE_CHANNELS
    channel_to_check = level;
#else
    channel_to_check = DEBUG_CHANNEL;
#endif

    // swo_util_is_channel_active already checks if channel_to_check < 32
    bool channel_active = swo_util_is_channel_active(channel_to_check);

#if USE_CHANNEL_MASK_VARIABLE
    return channel_active && ((MASK_VARIABLE & (1UL << level)) != 0UL);
#else
    #if !USE_MULTIPLE_CHANNELS && !USE_CHANNEL_MASK_VARIABLE
        return (level <= LOG_LEVEL_NOISY) && channel_active;
    #else
        return channel_active;
    #endif
#endif
}

/* Write a message to the SWO backend */
static void swo_backend_write(const log_message_t* message) {
    if (message == NULL) { // Corrected null check
        return;
    }
    // Assuming message->message is ensured to be non-NULL by the logger core if message itself is not NULL.
    // Or, if message->message can be NULL, the caller of swo_util_puts should handle it or swo_util_puts should be robust to it.

#if USE_MULTIPLE_CHANNELS
    swo_util_puts(message->message, message->level);
#else
    swo_util_puts(message->message, DEBUG_CHANNEL);
#endif
}

/* Flush the SWO backend - not needed for SWO */
static void swo_backend_flush(void) {
    /* SWO is a real-time interface, no buffering, nothing to flush */
}

/* Clear the console for SWO */
static void swo_backend_clear(void) {
#if USE_ANSI_ESCAPE_CODES
    swo_util_puts("\033[2J\033[H", DEBUG_CHANNEL); // ANSI escape code to clear screen and move cursor to top-left
#endif
}